grammar datalog;

/********************************************************
 * Productions
 ********************************************************/
datalogProgram	:	SCHEMES COLON (schemeList)
		        FACTS COLON (factList | importFacts)
		        RULES COLON (infRulesList | importInfRules)
		        POSITIVE COLON (posRulesList | importPosRules)
		        NEGATIVE COLON (negRulesList | importNegRules)
		        KEY COLON (keyRulesList | importKeyRules)
		        PROGRAM COLON queryList
		        //QUERIES COLON (query | aggQuery)
                ;

scheme   	: 	ID LEFT_PAREN idList RIGHT_PAREN PERIOD;

schemeList	:	(scheme)+;

//importScheme   	:   SHELL IMPORT FROM DIR d=STRING;


fact    	: 	ID LEFT_PAREN valueList RIGHT_PAREN PERIOD;

factList	:	fact+;

importFacts   	:   SHELL IMPORT FROM MYSQL db=STRING u=STRING mdp=STRING;

ruleA    	:	atom COLON_DASH atomList PERIOD;

infRulesList	:	ruleA*;

importInfRules   	:   SHELL IMPORT FROM DIR d=STRING;

posRulesList	:	ruleA*;

importPosRules   	:   SHELL IMPORT FROM DIR d=STRING;

negRuleA    	:	NEG COLON_DASH atomList PERIOD;

negRulesList	:	negRuleA*;

importNegRules   	:   SHELL IMPORT FROM DIR d=STRING;

keyRulesList	:	ruleA*;

importKeyRules   	:   SHELL IMPORT FROM DIR d=STRING;

//headAtom	:	ID LEFT_PAREN elementList RIGHT_PAREN
//                    ;

aggQueryHeadAtom	:	pd=ID LEFT_PAREN elementList COMMA ID RIGHT_PAREN
                    ;

atom	:	ID LEFT_PAREN elementList RIGHT_PAREN
                ;

atomList	:	(atom COMMA)* atom
                    ;

element	:	ID | STRING | INT | FLOAT
                ;

elementList	: 	(element COMMA)* element
                    ;

idList  	:	(ID COMMA)* ID;

query           : simpleQuery | aggQuery | mapQuery | betaQuery
                ;

simpleQuery	    :   Q_MARK? atom COLON_DASH atomList (COMMA assigExpression)? PERIOD
                ;

aggQuery	        :      Q_MARK? aggQueryHeadAtom COLON_DASH
                              AGG LEFT_PAREN atomList (COMMA idList)? COMMA assigAggr RIGHT_PAREN PERIOD
                        ;

mapQuery	        :      Q_MARK? atom COLON_DASH
                              MAP LEFT_PAREN nodes=atom COMMA links=atom COMMA assigExpression RIGHT_PAREN PERIOD
                        ;

betaQuery	        :      Q_MARK? head=atom COLON_DASH
                              BETA LEFT_PAREN nodes=atom COMMA links=atom COMMA n=INT (COMMA idList)? COMMA map=assigExpression COMMA reduce=assigAggr COMMA update=assigExpression RIGHT_PAREN PERIOD
                        ;

//betaQuery	        :      Q_MARK? aggQueryHeadAtom COLON_DASH
//                              BETA LEFT_PAREN r=ID COMMA v=ID COMMA e=ID COMMA n=INT COMMA dir=(IN|OUT|BOTH) COMMA '[' follow=idList ']' COMMA '{' set=assigExpression '}' COMMA '{' map=assigExpression '}' COMMA '{' reduce=assig_aggr '}' COMMA '{' update=assigExpression '}' RIGHT_PAREN PERIOD
//                              ;


queryList	:	(query)+
                ;

assigExpression: variable=(ID|QUALIFIED_ID) ATTRIBUTION expr;

assigAggr: variable=ID ATTRIBUTION func=(MIN|MAX|SUM|COUNT) LEFT_PAREN val=ID RIGHT_PAREN;

assig_aggr: attribute=(ID|QUALIFIED_ID) '=' aggr=(MIN|MAX|SUM)'(' expr ')' 'BY' '[' idList ']';

list_identifiers: (QUALIFIED_ID ',')* QUALIFIED_ID;

expr:	expr op=('*'|'/') expr                       # expr_md
    |	expr op=('+'|'-') expr                       # expr_pm
    |	INT                                          # expr_int
    |	FLOAT                                        # expr_float
    |   ID                                           # expr_var
    |   table=QUALIFIED_ID                           # expr_tableid
    |   aggr=(MIN|MAX|SUM) '(' '[' list_identifiers ']' ')'       # exp_agg
    |	'(' expr ')'                                 # expr_parens
    ;


valueList	: 	((STRING | INT | FLOAT) COMMA)* (STRING | INT | FLOAT);

/********************************************************
 * Token definitions 
 ********************************************************/

COMMA 	:	','
	;

PERIOD	:	'.'
	;

Q_MARK	:	'?'
	;

LEFT_PAREN	: '('
		;

RIGHT_PAREN	: ')'
		;

COLON		: ':'
		;

COLON_DASH	: ':-' | '-:'
		;

ATTRIBUTION	: '='
		;

MULTIPLY	: '*'
		;

ADD		: '+'
		;

NEG		: '!'
		;

SHELL		: '$'
     		;

AGG		: 'aggr'
     		;

MAP		: 'map'
     		;

BETA		: 'beta'
     		;

MIN   : 'min';

MAX   :  'max';

SUM   :  'sum';

COUNT   :  'count';

SCHEMES		: 'Schemes'
		;
FACTS		: 'Facts'
		;
RULES		: 'Inference Rules'
		;
POSITIVE		: 'Positive Constraints'
		;
NEGATIVE		: 'Negative Constraints'
		;
KEY		: 'Key Constraints'
		;
PROGRAM		: 'Program'
		;
IMPORT		: 'IMPORT'
      		;
FROM		: 'FROM'
    		;
DIR		: 'DIR'
    		;
MYSQL		: 'MYSQL'
    		;

IN : 'IN';
OUT  : 'OUT';
BOTH  : 'BOTH';

VTABLE: 'V';
ETABLE: 'E';
NEW: 'newV';
CURRENT: 'current';

ID  :	('a'..'z'|'A'..'Z') ('a'..'z'|'A'..'Z'|'0'..'9'|'_')*
    ;

QUALIFIED_ID: (VTABLE|ETABLE|NEW|CURRENT)'.'ID;

STRING
    :  '\'' ( ESC_SEQ | ~('\\'|'\'') )* '\''
    ;

FLOAT: DIGIT+ '.' DIGIT+;

INT     : DIGIT+ ;

COMMENT
    :   '%' ~[\r\n]* -> skip
    ;

WS  :   [ \t\r\n]+ -> skip
    ;

fragment
DIGIT : [0-9] ;

fragment
HEX_DIGIT : ('0'..'9'|'a'..'f'|'A'..'F') ;

fragment
ESC_SEQ
    :   '\\' ('b'|'t'|'n'|'f'|'r'|'\"'|'\''|'\\')
    |   UNICODE_ESC
    |   OCTAL_ESC
    ;

fragment
OCTAL_ESC
    :   '\\' ('0'..'3') ('0'..'7') ('0'..'7')
    |   '\\' ('0'..'7') ('0'..'7')
    |   '\\' ('0'..'7')
    ;

fragment
UNICODE_ESC
    :   '\\' 'u' HEX_DIGIT HEX_DIGIT HEX_DIGIT HEX_DIGIT
    ;